@@ -1,5 +1,5 @@
 import copy
-import itertools
+import gc
 import json
 import logging
 
@@ -18,11 +18,9 @@ IS_DUMMY = "_is_dummy"
 
 
 def assert_2_layer_list(x):
-    assert isinstance(x, list) and all(isinstance(i, list) for i in x), (
-        "For parquet, AncDataset's filepaths should be a 2-layer list: list of datasets (ParquetConcateDataset) -> "
-        "list of ParquetDataset. ParquetDataset accepts both a single str filepath and a list of filepaths for cases "
-        "including correlated files such as wav"
-    )
+    assert isinstance(x, list) and all(
+        isinstance(i, list) for i in x
+    ), "AncDataset's filepaths should be a 2-layer list: list of datasets -> list of files in the dataset."
 
 
 def get_sub_ds_info(sub_ds_info_filepath, parquet_filepaths):
@@ -80,10 +78,13 @@ class AncDataset(IterableDataset):
         enable_logging=True,
         state_queues=None,
         ckpt_interval=None,
+        skip_read_failures=False,
+        worker_gc_freq=None,
     ):
         assert isinstance(
             processor, Processor
         ), "processor must be an anc_processor.Processor"
+        assert_2_layer_list(filepaths)
         self.filepaths = filepaths
         self.dataset_type = dataset_type
         processor.ds_args = ds_args
@@ -91,15 +92,20 @@ class AncDataset(IterableDataset):
         decode_fn = processor.decode_fn if hasattr(processor, "decode_fn") else None
         self.ds = []
         if dataset_type == "parquet":
-            assert_2_layer_list(filepaths)
             # each filepath represents a dataset
             for idx, filepath in enumerate(filepaths):
+                ds_idx = idx
+                if ds_args.get("_loader_idx"):
+                    ds_idx = ds_args.get("_loader_idx")
                 fs_config = ds_args.get("file_system")
                 if ds_args.get("dataset_file_system"):
-                    fs_config = ds_args.get("dataset_file_system")[idx]
+                    assert isinstance(
+                        ds_args.get("dataset_file_system"), list
+                    ), "dataset_file_system should be a list"
+                    fs_config = ds_args.get("dataset_file_system")[ds_idx]
                 sub_ds_info_config = None
-                if ds_args.get("sub_ds_info_filepath"):
-                    sub_ds_info_config = ds_args.get("sub_ds_info_filepath")[idx]
+                if isinstance(ds_args.get("sub_ds_info_filepath"), list):
+                    sub_ds_info_config = ds_args.get("sub_ds_info_filepath")[ds_idx]
                 self.ds.append(
                     ParquetConcateDataset(
                         filepath,
@@ -109,15 +115,12 @@ class AncDataset(IterableDataset):
                     )
                 )
         elif dataset_type == "jsonl":
-            assert isinstance(
-                filepaths[0], str
-            ), "jsonl dataset does not support multisources"
-            assert len(filepaths) == 1, (
-                f"jsonl dataset only support single file, but got {len(filepaths)}. "
-                "Please convert jsonl dataset to parquet first"
+            assert len(filepaths) == 1 and len(filepaths[0]) == 1, (
+                f"jsonl dataset only support single file and single dataset, but got "
+                f"{len(filepaths)} datasets and {len(filepaths[0])} files."
             )
             # TODO: support file_system for jsonl dataset
-            self.ds = [JsonlDataset(filepaths[0], decode_fn=decode_fn)]
+            self.ds = [JsonlDataset(filepaths[0][0], decode_fn=decode_fn)]
         self.processor = processor
         self.sampler = sampler
         self.enable_compose = enable_compose
@@ -146,6 +149,10 @@ class AncDataset(IterableDataset):
         self.ds_states = None
         self.remain_data = []
         self.raw_data = []
+        self.skip_read_failures = skip_read_failures
+        if worker_gc_freq is not None:
+            assert worker_gc_freq > 0, "worker_gc_freq should be greater than 0"
+        self.worker_gc_freq = worker_gc_freq
 
     def set_sampler(self, sampler):
         self.sampler = sampler
@@ -153,16 +160,19 @@ class AncDataset(IterableDataset):
     def __iter__(self):
         worker_info = torch.utils.data.get_worker_info()
         wid = worker_info.id if worker_info is not None else 0
+        if wid == 0:
+            if self.worker_gc_freq is None:
+                logging.info("worker_gc_freq is None, using python's default gc")
+                gc.enable()
+            else:
+                logging.info(f"worker_gc_freq is every {self.worker_gc_freq} batches")
         num_workers = worker_info.num_workers if worker_info is not None else 1
         if self.ds_states is not None:
             self._set_ckpt(self.ds_states[wid])
         logging.debug(
             f"iter information: len(sampler) - {len(self.sampler)}, worker id - {wid}, num_workers - {num_workers}"
         )
-        # use islice to get the iterator of sampler only for current worker
-        sampler_iter = itertools.islice(
-            iter(self.sampler), wid, len(self.sampler), num_workers
-        )
+        sampler_iter = iter(self.sampler)
         yield from self.get_generator(sampler_iter, wid, num_workers)
 
     def get_generator(self, sampler_iter, wid, num_workers):
@@ -183,6 +193,11 @@ class AncDataset(IterableDataset):
                 yield (wid, result, False)
 
                 self.cur_step += 1
+                if (
+                    self.worker_gc_freq is not None
+                    and self.cur_step % self.worker_gc_freq == 0
+                ):
+                    gc.collect()
                 if (
                     self.ckpt_interval is not None
                     and self.cur_step % self.ckpt_interval == 0
@@ -204,7 +219,15 @@ class AncDataset(IterableDataset):
             return
         new_data = []
         for idx in idxs:
-            sample = self.ds[ds_idx][idx]
+            try:
+                sample = self.ds[ds_idx][idx]
+            except Exception as e:
+                if not self.skip_read_failures:
+                    raise e
+                logging.warning(
+                    f"Failed to get sample from dataset {ds_idx} at index {idx}: {e}, skipping this sample"
+                )
+                continue
             # TODO: create a reserved key list for "data", add _ds_idx and filepath to the list
             if (
                 isinstance(sample, dict)
