@@ -1,10 +1,13 @@
 import bisect
 import copy
+import heapq
 import logging
 import math
 import random
+import time
 
 import numpy as np
+import torch
 from torch.utils.data.distributed import DistributedSampler
 
 from anc_data.utils import EndOfIteration
@@ -15,7 +18,7 @@ class CustomizedRange:
         self.start = start
         self.end = end
         self._seed = seed
-        self.reset_random_generator()
+        self._random = None
 
     def __len__(self):
         return self.end - self.start
@@ -27,6 +30,8 @@ class CustomizedRange:
         return CustomizedRange(self.start, self.end, self._seed + clone_times)
 
     def to_list(self, shuffle=False):
+        if self._random is None:
+            self.reset_random_generator()
         lst = np.arange(self.start, self.end, dtype=np.uint32)
         if shuffle:
             self._random.shuffle(lst)
@@ -50,11 +55,13 @@ class CustomizedRange:
 
 
 class RangesHolder:
-    def __init__(self, list_of_ranges):
+    def __init__(self, list_of_ranges, expand_size_limit=20):
         self.list_of_ranges = list_of_ranges
         self.accu_lengths = self.create_accu_length(list_of_ranges)
         self.length = sum(len(r) for r in list_of_ranges)
         self.expanded = {}
+        self.expand_size_limit = expand_size_limit
+        self.expanded_heap = []  # (time, range_idx)
         self.i = 0
 
     def __len__(self):
@@ -120,7 +127,7 @@ class RangesHolder:
                 res.append(self.list_of_ranges[i].slice(0, end_range_offset))
             else:
                 res.append(self.list_of_ranges[i])
-        target = RangesHolder(res)
+        target = RangesHolder(res, self.expand_size_limit)
         assert target.length == length
         return target
 
@@ -142,10 +149,31 @@ class RangesHolder:
     def expand(self, range_idx, shuffle=False):
         if range_idx in self.expanded:
             return self.expanded[range_idx]
+        elif self.expand_size_limit > 0:
+            self.expanded[range_idx] = self.list_of_ranges[range_idx].to_list(shuffle)
+            if len(self.expanded_heap) < self.expand_size_limit:
+                heapq.heappush(self.expanded_heap, (time.time(), range_idx))
+            else:
+                _, range_idx_to_shrink = heapq.heappushpop(
+                    self.expanded_heap, (time.time(), range_idx)
+                )
+                assert range_idx_to_shrink != range_idx
+                self.shrink(range_idx_to_shrink)
+            assert len(self.expanded) == len(
+                self.expanded_heap
+            ), f"len(self.expanded) {len(self.expanded)} != len(self.expanded_heap) {len(self.expanded_heap)}"
+            return self.expanded[range_idx]
         else:
             self.expanded[range_idx] = self.list_of_ranges[range_idx].to_list(shuffle)
             return self.expanded[range_idx]
 
+    def shrink(self, range_idx):
+        assert (
+            range_idx in self.expanded
+        ), f"range idx {range_idx} is not found in self.expanded"
+        self.expanded.pop(range_idx)
+        self.list_of_ranges[range_idx]._random = None
+
     def get_value_from(self, start, end, shuffle=False):
         start_range_id, start_range_offset, end_range_id, end_range_offset = (
             self.get_start_and_end_ranges(start, end - start)
@@ -175,6 +203,7 @@ class RangesHolder:
 
     def random_reset(self, idx, offset=1):
         self.expanded = {}
+        self.expanded_heap = []
         self.list_of_ranges = self.list_of_ranges[idx:] + self.list_of_ranges[:idx]
         self.accu_lengths = self.create_accu_length(self.list_of_ranges)
         for range in self.list_of_ranges:
@@ -228,6 +257,7 @@ class AncSampler(DistributedSampler):
         chunk_granularity=-1,
         mem_save_mode=False,
         max_samples=None,
+        range_expand_size_limit=20,
     ):
         if hasattr(dataset, "get_sub_lengths"):
             self.sub_lengths = dataset.get_sub_lengths()
@@ -251,6 +281,7 @@ class AncSampler(DistributedSampler):
         self.drop_last = drop_last
         self.chunk_granularity = chunk_granularity
         self.mem_save_mode = mem_save_mode
+        self.range_expand_size_limit = range_expand_size_limit
         if self.mem_save_mode:
             assert (
                 self.chunk_granularity == -1
@@ -406,40 +437,42 @@ class AncSampler(DistributedSampler):
             start_idx_per_worker[i] = (
                 start_idx_per_worker[i - 1] + num_samples_per_worker[i - 1]
             )
-        return num_batches, num_samples_per_worker, start_idx_per_worker
+        return num_samples_per_worker, start_idx_per_worker, num_batches_per_each_worker
 
     def _create_chunk_indices_from_bz_and_worker(self, indices):
-        num_batches, num_samples_per_worker, start_idx_per_worker = (
+        num_samples_per_worker, start_idx_per_worker, num_batches_per_each_worker = (
             self._compute_worker_indices_ranges()
         )
-
+        worker_info = torch.utils.data.get_worker_info()
+        self.wid = worker_info.id if worker_info is not None else 0
         if self.repeat and self.shuffle:
             self._np_random = np.random.default_rng(self.seed)
         while True:
             self.inner_epoch_count += 1
-            current_idx_per_worker = [i for i in start_idx_per_worker]
-            for i in range(num_batches):
-                wid = i % self.num_workers
-                start_idx = current_idx_per_worker[wid]
+
+            start_idx = start_idx_per_worker[self.wid]
+            for i in range(num_batches_per_each_worker[self.wid]):
                 end_idx = start_idx + self.batch_size
                 is_last_batch = (
-                    end_idx - start_idx_per_worker[wid]
-                ) >= num_samples_per_worker[wid] and not self.repeat
+                    end_idx - start_idx_per_worker[self.wid]
+                ) >= num_samples_per_worker[self.wid] and not self.repeat
                 end_idx = min(
                     end_idx,
-                    start_idx_per_worker[wid] + num_samples_per_worker[wid],
+                    start_idx_per_worker[self.wid] + num_samples_per_worker[self.wid],
                 )
                 if self.resume_step == 0:
-                    # done skipping
                     if self.mem_save_mode:
-                        yield indices.get_value_from(
-                            start_idx, end_idx, self.shuffle
-                        ).tolist(), is_last_batch
+                        yield (
+                            indices.get_value_from(
+                                start_idx, end_idx, self.shuffle
+                            ).tolist(),
+                            is_last_batch,
+                        )
                     else:
                         yield indices[start_idx:end_idx].tolist(), is_last_batch
                 else:
                     self.resume_step -= 1
-                current_idx_per_worker[wid] = end_idx
+                start_idx = end_idx
 
             if not self.repeat:
                 break
@@ -484,7 +517,9 @@ class AncSampler(DistributedSampler):
                     else:
                         indices = np.concatenate(sub_indices_list)
                 else:
-                    indices = RangesHolder(sub_indices_list)
+                    indices = RangesHolder(
+                        sub_indices_list, self.range_expand_size_limit
+                    )
             else:
                 if not self.mem_save_mode:
                     indices = self._get_indices(self.sub_lengths, ratio=self.ratios)
@@ -497,7 +532,9 @@ class AncSampler(DistributedSampler):
                     )
                     if self.shuffle:
                         self._np_random.shuffle(sub_indices_list)
-                    indices = RangesHolder(sub_indices_list)
+                    indices = RangesHolder(
+                        sub_indices_list, self.range_expand_size_limit
+                    )
 
             # if max_samples is set, may need to truncate indices to total_size
             if self.drop_last or len(indices) >= self.total_size:
@@ -557,13 +594,6 @@ class AncSampler(DistributedSampler):
         # python's variable scope is function based not block based
         indices_generator = self._iter_init()
 
-        # yield empty indices to let workers follow the original order
-        # otherwise, the first generated indices would be consumed by worker 0
-        # while they should be consumed by cur_wid
-        cur_wid = self.resume_step % self.num_workers
-        for i in range(cur_wid):
-            yield [], False
-
         while True:
             self.step += 1
             indices = next(indices_generator, EndOfIteration)
@@ -617,11 +647,18 @@ class AncMultiSourceSampler:
         global_shuffle=False,
         chunk_granularity=-1,
         mem_save_mode=False,
+        sequential=False,
     ):
         self.ratios = ratios
         self.seed = seed + rank
-        assert repeat is True
-        assert 0.9999 <= sum(ratios) <= 1.0001
+        if not sequential:
+            assert 0.9999 <= sum(ratios) <= 1.0001
+        self.sequential = sequential
+        if sequential and len(ratios) > 1:
+            assert len(ratios) == len(
+                dataset
+            ), "ratios must be the same length as dataset when sequential is True"
+            logging.info(f"each dataset will repeat according to the ratios: {ratios}")
         self.samplers = [
             AncSampler(
                 dataset.get_sub_lengths(i),
@@ -637,35 +674,50 @@ class AncMultiSourceSampler:
                 global_shuffle=global_shuffle,
                 chunk_granularity=chunk_granularity,
                 mem_save_mode=mem_save_mode,
+                ratios=[self.ratios[i]] if sequential and len(self.ratios) > 1 else [1],
             )
             for i in range(len(dataset))
         ]
+        self.repeat = repeat
         self.random = random.Random(self.seed)
-        self.shuffle = shuffle
+        if self.repeat and self.sequential:
+            # TODO: implement
+            raise NotImplementedError(
+                "repeat and sequential are not supported together"
+            )
+        self.cur_iter_idx = 0
 
     def __iter__(self):
         iters = [iter(sampler) for sampler in self.samplers]
         iter_list = list(range(len(iters)))
-        if not self.shuffle:
-            # if shuffle is False, we want every epoch sampler output the same indices
-            self.random = random.Random(self.seed)
         while True:
-            cur_iter_idx = self.random.choices(iter_list, weights=self.ratios)[0]
-            cur_iter = iters[cur_iter_idx]
-            indices, is_last_batch = next(cur_iter)
-            yield cur_iter_idx, indices, is_last_batch
+            if not self.sequential:
+                self.cur_iter_idx = self.random.choices(iter_list, weights=self.ratios)[
+                    0
+                ]
+            cur_iter = iters[self.cur_iter_idx]
+            result = next(cur_iter, EndOfIteration)
+            if result is EndOfIteration:
+                self.cur_iter_idx += 1
+                if self.cur_iter_idx >= len(iters):
+                    break
+                continue
+            indices, is_last_batch = result
+
+            yield self.cur_iter_idx, indices, is_last_batch
 
     def is_padding(self, ds_idx, index):
         return self.samplers[ds_idx].is_padding(ds_idx, index)
 
     def __len__(self):
-        return 1000000000000
+        return int(np.sum(len(sampler) for sampler in self.samplers))
 
     def _get_ckpt(self):
         state = {}
         samplers_state = [sampler._get_ckpt() for sampler in self.samplers]
         state["samplers"] = samplers_state
         state["random_state"] = self.random.getstate()
+        state["cur_iter_idx"] = self.cur_iter_idx
         return state
 
     def _set_ckpt(self, state):
@@ -673,3 +725,4 @@ class AncMultiSourceSampler:
         for i, sampler in enumerate(self.samplers):
             sampler._set_ckpt(state["samplers"][i])
         self.random.setstate(state["random_state"])
+        self.cur_iter_idx = state["cur_iter_idx"]
