@@ -23,7 +23,13 @@ class JsonlDataset:
         assert (
             0 <= idx < len(self.data)
         ), f"idx shoud be in [0, {len(self.data)}) but get {idx}"
-        return self.data[idx]
+        item = self.data[idx]
+        if isinstance(item, dict):
+            if "filepath" not in item:
+                item["filepath"] = self.filepath
+            if "row_idx" not in item:
+                item["row_idx"] = idx
+        return item
 
     def get_sub_lengths(self, level="file"):
         return [len(self.data)]
