@@ -157,7 +157,7 @@ class ParquetDataset:
         for idx, pf in enumerate(self.parquet_files):
             assert self.num_row_groups == pf.num_row_groups, (
                 f"parquet file {self.filepath[idx]} has {pf.num_row_groups} row groups but "
-                "{self.filepath[0]} has {self.num_row_groups}"
+                f"{self.filepath[0]} has {self.num_row_groups}"
             )
             assert self.row_group_rows == [
                 pf.metadata.row_group(rid).num_rows for rid in self.row_groups
@@ -204,10 +204,7 @@ class ParquetDataset:
             yield from rows
 
     def _clean_cache(self):
-        # TODO: maybe a better way to find out which row groups to delete
-        gidx_to_delete = [i for i in self.cache]
-        for i in gidx_to_delete:
-            del self.cache[i]
+        self.cache.clear()
 
     def __getitem__(self, idx):
         self.__iter__init__()
@@ -307,10 +304,9 @@ class ParquetConcateDataset:
         return self.num_rows
 
     def _clean_cache(self):
-        ds_idx_to_delete = [i for i in self.cache]
-        for i in ds_idx_to_delete:
+        for i in self.cache.keys():
             self.cache[i].close()
-            del self.cache[i]
+        self.cache.clear()
 
     def __getitem__(self, idx):
         if idx < 0:
