@@ -81,6 +81,7 @@ class PretrainProcessor(Processor):
         special_token_ids=None,
         pad_to_max_seq_len=False,
         seq_split_config=None,
+        remove_special_tokens=True,
     ):
         self.opt = FakeOpt(hf_model_name, "", add_bos)
         self.tokenizer = HFPretrainedTokenizer(self.opt)
@@ -97,6 +98,7 @@ class PretrainProcessor(Processor):
         self.pad_token_id = self.tokenizer.end_token_id
         self.cu_seqlen_use_old_pack = cu_seqlen_use_old_pack
         self.pad_to_max_seq_len = pad_to_max_seq_len
+        self.remove_special_tokens = remove_special_tokens
         if special_token_ids is None:
             special_token_ids = []
         if special_tokens is not None:
@@ -173,8 +175,10 @@ class PretrainProcessor(Processor):
         # may need more complex logic when we have more special tokens, or mask logic becomes more complex
         set_to_zero = False
         successive_zero_count = 0
+        special_token_idxs = []
         for i in range(len(tokens)):
             if tokens[i] in self.special_token_ids:
+                special_token_idxs.append(i)
                 set_to_zero = not set_to_zero
                 loss_mask[i] = 0
                 successive_zero_count += 1
@@ -197,6 +201,33 @@ class PretrainProcessor(Processor):
                     return None
             else:
                 successive_zero_count = 0
+        # check if the first special token is the mask end token
+        if special_token_idxs and self.tokenizer.vec2txt(
+            tokens[special_token_idxs[0]]
+        ).startswith("</"):
+            # flip the loss mask
+            logging.warn(
+                f"Flipping loss mask for sample from {item.get('filepath', 'unknown')}:{item.get('row_idx', 'unknown')}"
+            )
+            special_token_idxs = set(special_token_idxs)
+            for i in range(len(loss_mask)):
+                if i not in special_token_idxs:
+                    loss_mask[i] = 1 - loss_mask[i]
+
+        # to remove special tokens from the input_ids
+        if self.remove_special_tokens and special_token_idxs:
+            new_tokens = [
+                tokens[i] for i in range(len(tokens)) if i not in special_token_idxs
+            ]
+            new_loss_mask = [
+                loss_mask[i]
+                for i in range(len(loss_mask))
+                if i not in special_token_idxs
+            ]
+            tokens = new_tokens
+            loss_mask = new_loss_mask
+            assert len(tokens) == len(loss_mask)
+
         if sum(loss_mask) == 0:
             self.all_masked_sample_cnt += 1
             if self.sample_cnt >= 1000:
